/************************************************************************
 *   IRC - Internet Relay Chat, modules/m_accept.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "client.h"
#include "ircd.h"
#include "numeric.h"
#include "s_conf.h"
#include "s_serv.h"
#include "send.h"
#include "msg.h"
#include "parse.h"
#include "config.h"
#include "m_commands.h"
#include "tools.h"

#ifdef UMODE_G

/*
 * accept_message
 *
 * inputs       - pointer to source client
 *              - pointer to target client
 * output       - 1 if accept this message 0 if not
 * side effects - See if source is on target's allow list
 */
int accept_message(struct Client *source, struct Client *target)
{
  dlink_node *ptr;
  struct Client *target_p;
 
  for(ptr = target->allow_list.head; ptr; ptr = ptr->next )
    {
      target_p = ptr->data;
      if(source == target_p)
        return 1;
    }
  return 0;
}

/*
 * add_to_accept
 *
 * inputs       - pointer to source client
 *              - pointer to target client
 * output       -
 * side effects - Add's source pointer to targets allow list
 */
void add_to_accept(struct Client *source, struct Client *target)
{
  dlink_node *m;
  int len;
 
  /* Safety checks, neither of these tests should happen */
  if (!IsPerson(source))
    return;
 
  if (!IsPerson(target))
    return;
 
  /* XXX MAX_ALLOW should be in config file not hard coded */
  if ( (len = dlink_list_length(&target->allow_list)) >=
       MAX_ALLOW)
    {
      sendto_one(target,":%s NOTICE %s :Max accept targets reached %d",
                 me.name, target->name, len);
      return;
    }
 
  m = make_dlink_node();
  dlinkAdd(source, m, &target->allow_list);
 
  m = make_dlink_node();
  dlinkAdd(target, m, &source->on_allow_list);
}

/*
 * del_from_accept
 *
 * inputs       - pointer to source client
 *              - pointer to target client
 * output       - NONE
 * side effects - Deletes source pointer to targets allow list
 *
 * Walk through the target's accept list, remove if source is found,
 * Then walk through the source's on_accept_list remove target if found.
 */
void del_from_accept(struct Client *source, struct Client *target)
{
  dlink_node *ptr;
  dlink_node *ptr2;
  dlink_node *next_ptr;
  dlink_node *next_ptr2;
  struct Client *target_p;
 
  for (ptr = target->allow_list.head; ptr; ptr = next_ptr)
    {
      next_ptr = ptr->next;
 
      target_p = ptr->data;
      if(source == target_p)
        {
          dlinkDelete(ptr, &target->allow_list);
          free_dlink_node(ptr);
 
          for (ptr2 = source->on_allow_list.head; ptr2;
               ptr2 = next_ptr2)
            {
              next_ptr2 = ptr2->next;
 
              target_p = ptr2->data;
              if (target == target_p)
                {
                  dlinkDelete(ptr2, &source->on_allow_list);
                  free_dlink_node(ptr2);
                }
            }
        }
    }
}

/*
 * del_all_accepts
 *
 * inputs       - pointer to exiting client
 * output       - NONE
 * side effects - Walk through given clients allow_list and on_allow_list
 *                remove all references to this client
 */
void del_all_accepts(struct Client *client_p)
{
  dlink_node *ptr;
  dlink_node *next_ptr;
  struct Client *target_p;
 
  for (ptr = client_p->allow_list.head; ptr; ptr = next_ptr)
    {
      next_ptr = ptr->next;
      target_p = ptr->data;
      if(target_p != NULL)
        del_from_accept(target_p,client_p);
    }
 
  for (ptr = client_p->on_allow_list.head; ptr; ptr = next_ptr)
    {
      next_ptr = ptr->next;
      target_p = ptr->data;
      if(target_p != NULL)
        del_from_accept(client_p, target_p);
    }
}

/*
 * list_all_accepts
 *
 * inputs       - pointer to exiting client
 * output       - NONE
 * side effects - list allow list
 */
void list_all_accepts(struct Client *source_p)
{
  dlink_node *ptr;
  struct Client *target_p;
  char *nicks[8];
  int j=0;
 
  nicks[0] = nicks[1] = nicks[2] = nicks[3] = nicks[4] = nicks[5]
    = nicks[6] = nicks[7] = "";
 
  sendto_one(source_p,":%s NOTICE %s :*** Current accept list",
             me.name, source_p->name);
 
  for (ptr = source_p->allow_list.head; ptr; ptr = ptr->next)
    {
      target_p = ptr->data;
 
      if(target_p != NULL)
        {
          nicks[j++] = target_p->name;
        }
 
      if(j > 7)
        {
          sendto_one(source_p,":%s NOTICE %s :%s %s %s %s %s %s %s %s",
                     me.name, source_p->name,
                     nicks[0], nicks[1], nicks[2], nicks[3],
                     nicks[4], nicks[5], nicks[6], nicks[7] );
          j = 0;
          nicks[0] = nicks[1] = nicks[2] = nicks[3] = nicks[4] = nicks[5]
            = nicks[6] = nicks[7] = "";
        }
 
    }

  if(j)
    sendto_one(source_p,":%s NOTICE %s :%s %s %s %s %s %s %s %s",
               me.name, source_p->name,
               nicks[0], nicks[1], nicks[2], nicks[3],
               nicks[4], nicks[5], nicks[6], nicks[7] );
}

/*
 * m_accept - ACCEPT command handler
 *      parv[0] = sender prefix
 *      parv[1] = servername
 */
int m_accept(struct Client *client_p, struct Client *source_p,
                    int parc, char *parv[])
{
  char *nick;
  int  add=1;
  struct Client *source;

  if (parc < 2) 
    {
      sendto_one(source_p, form_str(ERR_NEEDMOREPARAMS),
                 me.name, source_p->name, "ACCEPT");
      return 0;
    }

  nick = parv[1];

  add = 1;

  if (*nick == '-')
    {
      add = -1;
      nick++;
    }
  else if(*nick == '*')
    {
      list_all_accepts(source_p);
      return 0;
    }

  if ((source = find_client(nick,NULL)) == NULL)
    {
      sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name, parv[0], nick);
      return 0;
    }

  if (!IsPerson(source))
    {
      sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name, parv[0], nick);
      return 0;
    }

  if (add == 1)
    {
      if (accept_message(source, source_p)) {
	/* already on list */
	sendto_one(source_p, ":%s NOTICE %s :%s is already on your accept list",
		   me.name, parv[0], source->name);
	return 0;
      }
		
      del_from_accept(source,source_p);
      add_to_accept(source,source_p);
      sendto_one(source_p, ":%s NOTICE %s :Now allowing %s", 
		 me.name, parv[0], source->name);
      sendto_one(source, ":%s NOTICE %s :*** I've put you on my accept list.",
		      source_p->name, source->name);
    }
  else if(add == -1)
    {
		if (!accept_message(source, source_p)) 
		{
			/* not on list */
			sendto_one(source_p, ":%s NOTICE %s :%s is not on your accept list",
					   me.name, parv[0], source->name);
			return 0;
		}
		
      del_from_accept(source,source_p);
      sendto_one(source_p, ":%s NOTICE %s :Now removed %s from allow list", 
		 me.name, parv[0], source->name);
      sendto_one(source, ":%s NOTICE %s :*** I've taken you off my accept list.",
		      source_p->name, source->name);
    }
  return 0;
}

#endif
